TopoARTAM <- function() {
	
#--------------------------------------------------------------------------------------------------
	
	setClass("TopoARTAMNode",
		representation(
			node_ID				=	"numeric",
			cluster_ID 			=	"numeric",
			dim					=	"numeric",
			represented_inputs	=	"numeric",
			activation			=	"numeric",
			match_value			=	"numeric",
			edge_num			=	"numeric",
			edge_targets		=	"vector",
			weights				=	"vector"
		),
		prototype(
			node_ID				=	0,
			cluster_ID			=	0,
			dim					=	0,
			represented_inputs	=	0,
			activation			=	0,
			match_value			=	0,
			edge_num			=	0,
			edge_targets 		=	vector(mode = "list", length = 0),
			weights				=	vector(mode = "list", length = 0)
		)
	)
	
	setMethod("initialize", "TopoARTAMNode", function(.Object, src_file, node_index)
	{
		#tmpstring<-sprintf("load TopoART-AM node %i", node_index - 1)
		#print(tmpstring)
		
		f <- src_file
		l <- readLines(f, n = 1)
		
		node_ID				<-	ReadNumeric(f, "node ID: %i");
		cluster_ID			<-	ReadNumeric(f, "cluster ID: %i");
		dim					<-	ReadNumeric(f, "input length: %i");
		represented_inputs	<-	ReadNumeric(f, "represented inputs: %i");
		activation			<-	ReadNumeric(f, "activation: %f");
		match_value			<-	ReadNumeric(f, "match value: %f");
		
		l <- readLines(f, n = 1)
		edge_num <- ReadNumeric(f, "edge number: %i");
		edge_targets <- vector(mode = "numeric", length = edge_num)
		l <- readLines(f, n = 1)
		substrings <- strsplit(l, split = " ")
		for(i in seq_len(edge_num))
			edge_targets[i] <- substrings[[1]][i + 2]
		#print(edge_num)
		#print(edge_targets)
		
		l <- readLines(f, n = 1)
		
		weights <- vector(mode = "numeric", length = dim)
		l <- readLines(f, n = 1)
		substrings <- strsplit(l, split = " ")
		for(i in seq_len(dim))
			weights[i] <- as.numeric(substrings[[1]][i])
		#print(weights)
		
		.Object@node_ID				<-	node_ID
		.Object@cluster_ID			<-	cluster_ID
		.Object@dim					<-	dim
		.Object@represented_inputs	<-	represented_inputs
		.Object@activation			<-	activation
		.Object@match_value			<-	match_value
		.Object@edge_num			<-	edge_num
		.Object@edge_targets		<-	edge_targets
		.Object@weights				<-	weights
		.Object
	})
	
#--------------------------------------------------------------------------------------------------
	
	setClass("TopoARTAMModule",
		representation(
			x_F1_dim			=	"numeric",
			x_F1				=	"vector",
			rho					=	"numeric",
			next_ID				=	"numeric",
			valid_cluster_IDs	=	"logical",
			cluster_num			=	"numeric",
			LearningCycles		=	"numeric",
			node_num			=	"numeric",
			nodes				=	"vector"
		),
		prototype(
			x_F1_dim			=	0,
			x_F1				=	vector(mode = "numeric", length = 0),
			rho					=	0,
			next_ID				=	0,
			valid_cluster_IDs	=	FALSE,
			cluster_num			=	0,
			LearningCycles		=	0,
			node_num			=	0,
			nodes				=	vector(mode = "list", length = 0)
		)
	)
	
	setMethod("initialize", "TopoARTAMModule", function(.Object, src_file, module_index)
	{
		letters <- c("a", "b", "c", "d", "e")
		tmpstring <- sprintf("load TopoART-AM module %s", letters[module_index])
		print(tmpstring)
		
		f <- src_file
		l <- readLines(f, n = 3)
		
		x_F1_dim <- ReadNumeric(f, "x^F1 length: %i");
		
		x_F1 <- vector(mode = "numeric", length = x_F1_dim)
		l <- readLines(f, n = 1)
		substrings <- strsplit(l, split = " ")
		for(i in seq_len(x_F1_dim))
			x_F1[i] <- substrings[[1]][i + 1]
		#print("x_F1:")
		#print(x_F1)
		
		rho		<-	ReadNumeric(f, "rho: %f");
		next_ID	<-	ReadNumeric(f, "next ID: %i");
		
		l <- readLines(f, n = 1)
		substrings <- strsplit(l, split = " ")
		logical_string <- substrings[[1]][4]
		valid_cluster_IDs <- (logical_string == "true")
		if(valid_cluster_IDs) {
			tmpstring2 <- "TRUE"
		} else {
			tmpstring2 <- "FALSE"
		}
		#tmpstring <- sprintf("valid_cluster_IDs: %s", tmpstring2)
		#print(tmpstring)
		
		cluster_num		<-	ReadNumeric(f, "cluster number: %i");
		LearningCycles	<-	ReadNumeric(f, "learning cycles: %i");
		node_num		<-	ReadNumeric(f, "node number: %i");	
		
		nodes <- vector(mode = "list", length = 0)	
		for(i in seq_len(node_num))
			nodes[[i]] <- new("TopoARTAMNode", f, i)
		
		.Object@x_F1_dim			<-	x_F1_dim
		.Object@x_F1				<-	x_F1
		.Object@rho					<-	rho
		.Object@next_ID				<-	next_ID
		.Object@valid_cluster_IDs	<-	valid_cluster_IDs
		.Object@cluster_num			<-	cluster_num
		.Object@LearningCycles		<-	LearningCycles
		.Object@node_num			<-	node_num
		.Object@nodes				<-	nodes
		.Object
	})
	
#--------------------------------------------------------------------------------------------------
	
	setClass("TopoARTAMNetwork", 
		representation(
			LibTopoART_version		=	"numeric",
			file_format_version		=	"numeric",
			TAM_file_format_version	=	"numeric",
			key1_dim				=	"numeric",
			key2_dim				=	"numeric",
			x_F0_dim				=	"numeric",
			x_F0					=	"vector",
			Rho_a					=	"numeric",
			Beta_sbm				=	"numeric",
			Tau						=	"numeric",
			LearningSteps			=	"numeric",
			Alpha					=	"numeric",
			ModuleNum				=	"numeric",
			Phis					=	"vector",
			modules					=	"vector"
		), 
		prototype(
			LibTopoART_version		=	0,
			file_format_version		=	0,
			TAM_file_format_version	=	0,
			key1_dim				=	0,
			key2_dim				=	0,
			x_F0_dim				=	0,
			x_F0					=	vector(mode = "numeric", length = 0),
			Rho_a					=	0,
			Beta_sbm				=	0,
			Tau						=	0,
			LearningSteps			=	0,
			Alpha					=	0,
			ModuleNum				=	0,
			Phis					=	vector(mode = "numeric", length = 0),
			modules					=	vector(mode = "list", length = 0)
		)
	)
	
	setMethod("initialize", "TopoARTAMNetwork", function(.Object, path)
	{
		tmpstring <- sprintf("load TopoART-AM network: %s", path)
		print(tmpstring)
		
		f <- file(path, "r")
		l <- readLines(f, n = 1)
		l <- readLines(f, n = 1)
		if(!(l == "*      TopoART-AM network       *")) {
			warning("No TopoART-AM text file")
			close(f)
			return(.Object)
		}
		
		l <- readLines(f, n = 1)
		substrings <- strsplit(l, split = "v")
		version_string <- substr(substrings[[1]][2], 1, 4)
		LibTopoART_version <- as.numeric(version_string)
		tmpstring <- sprintf("LibTopoART v%.2f", LibTopoART_version)
		print(tmpstring)
		l <- readLines(f, n = 1)
		
		l <- readLines(f, n = 1)
		substrings <- strsplit(l, split = " ")
		version_string <- substr(substrings[[1]][4], 1, 4)
		file_format_version <- as.numeric(version_string)
		version_string <- substr(substrings[[1]][5], 1, 4)
		TAM_file_format_version <- as.numeric(version_string)
		#tmpstring <- sprintf("file format versions: %.2f; %.2f", file_format_version, TAM_file_format_version)
		#print(tmpstring)
		if((file_format_version != 0.01) | ((TAM_file_format_version != 0.10) & (TAM_file_format_version != 0.11))) {
			warning("File format version not supported")
			close(f)
			return(.Object)
		}
		
		l <- readLines(f, n = 3)
		
		key1_dim <- ReadNumeric(f, "key_1 length: %i");
		key2_dim <- ReadNumeric(f, "key_2 length: %i");

		l <- readLines(f, n = 1)

		x_F0_dim <- ReadNumeric(f, "x^F0 length: %i")

		if(key1_dim + key2_dim != x_F0_dim) {
			warning("Input length mismatch")
			close(f)
			return(.Object)
		}
		
		x_F0 <- vector(mode = "numeric", length = x_F0_dim)
		l <- readLines(f, n = 1)
		substrings <- strsplit(l, split = " ")
		for(i in seq_len(x_F0_dim))
			x_F0[i] <- substrings[[1]][i + 1]
		#print("x^F0:")
		#print(x_F0)
		
		Rho_a			<-	ReadNumeric(f, "rho_a: %f");
		Beta_sbm		<-	ReadNumeric(f, "beta_sbm: %f");		
		Tau				<-	ReadNumeric(f, "tau: %i");			
		Phi				<-	ReadNumeric(f, "phi: %i");	
		LearningSteps	<-	ReadNumeric(f, "learning steps: %i");		
		Alpha			<-	ReadNumeric(f, "alpha: %f");	
		ModuleNum		<-	ReadNumeric(f, "module number: %i");
		
		Phis <- vector(mode = "numeric", length = ModuleNum)
		if(Phi != -1) {
			for(i in seq_len(ModuleNum))
				Phis[i] <- Phi
		} else {
			l <- readLines(f, n = 1)
			substrings <- strsplit(l, split = " ")
			for(i in seq_len(ModuleNum))
				Phis[i] <- as.numeric(substrings[[1]][i + 1])
		}

		modules <- vector(mode = "list", length = 0)	
		for(i in seq_len(ModuleNum))
			modules[[i]] <- new("TopoARTAMModule", f, i)
		
		close(f)
		
		.Object@LibTopoART_version		<-	LibTopoART_version
		.Object@file_format_version		<-	file_format_version
		.Object@TAM_file_format_version	<-	TAM_file_format_version
		.Object@key1_dim				<-	key1_dim
		.Object@key2_dim				<-	key2_dim
		.Object@x_F0_dim				<-	x_F0_dim
		.Object@x_F0					<-	x_F0
		.Object@Rho_a					<-	Rho_a
		.Object@Beta_sbm				<-	Beta_sbm
		.Object@Tau						<-	Tau
		.Object@LearningSteps			<-	LearningSteps
		.Object@Alpha					<-	Alpha
		.Object@ModuleNum				<-	ModuleNum
		.Object@Phis					<-	Phis
		.Object@modules					<-	modules
		.Object
	})
	
	setGeneric("SaveNetworkPlotKey1", function(this) standardGeneric("SaveNetworkPlotKey1"))
	setMethod("SaveNetworkPlotKey1", "TopoARTAMNetwork", function(this)
	{
		if((this@key1_dim != 2) || (this@key2_dim != 2)) {
			warning("Incompatible input length")
		} else {
			ColorShift <- 4
			MaxClusterNum <- max(this@modules[[this@ModuleNum]]@cluster_num, 2)
			colors <- rainbow(MaxClusterNum)
			colors[1] <- "#ffff80FF"
			colors[2] <- "#00A000FF"
			letters <- c("a", "b", "c", "d", "e")
			
			for(m in seq_len(this@ModuleNum)) {
				tmpstring <- sprintf("images/TAM_Key1_%s.pdf", letters[m])
				pdf(file = tmpstring)
				palette(colors)
				par(cex.axis = 1.5, cex.main = 1.6)
				plot(NULL, NULL, xlim = c(0, 1), ylim = c(0, 1), xlab = "", ylab = "")
				grid()
				tmpstring <- sprintf("TAM %s (key 1): rho_a=%.2f, beta_sbm=%.2f, phi=%i, tau=%i", letters[m], this@Rho_a, this@Beta_sbm, this@Phis[m], this@Tau)
				title(main = tmpstring, font.main = 2, cex.main = 1.45)

				for(n in seq_len(this@modules[[m]]@node_num)) {
					x1 <- this@modules[[m]]@nodes[[n]]@weights[[1]]
					y1 <- this@modules[[m]]@nodes[[n]]@weights[[2]]
					x2 <- 1 - this@modules[[m]]@nodes[[n]]@weights[[5]]
					y2 <- 1 - this@modules[[m]]@nodes[[n]]@weights[[6]]
					if(this@modules[[m]]@nodes[[n]]@cluster_ID > 0) {
						rect(x1, y1, x2, y2,col = (this@modules[[m]]@nodes[[n]]@cluster_ID - 1 + ColorShift) %% MaxClusterNum + 1)	
					}
				}
				
				dev.off()
			}
		}
	})

	setGeneric("SaveNetworkPlotKey2", function(this) standardGeneric("SaveNetworkPlotKey2"))
	setMethod("SaveNetworkPlotKey2", "TopoARTAMNetwork", function(this)
	{
		if((this@key1_dim != 2) || (this@key2_dim != 2)) {
			warning("Incompatible input length")
		} else {
			ColorShift <- 4
			MaxClusterNum <- max(this@modules[[this@ModuleNum]]@cluster_num, 2)
			colors <- rainbow(MaxClusterNum)
			colors[1] <- "#ffff80FF"
			colors[2] <- "#00A000FF"
			letters <- c("a", "b", "c", "d", "e")
			
			for(m in seq_len(this@ModuleNum)) {
				tmpstring <- sprintf("images/TAM_Key2_%s.pdf", letters[m])
				pdf(file = tmpstring)
				palette(colors)
				par(cex.axis = 1.5, cex.main = 1.6)
				plot(NULL, NULL, xlim = c(0, 1), ylim = c(0, 1), xlab = "", ylab = "")
				grid()
				tmpstring <- sprintf("TAM %s (key 2): rho_a=%.2f, beta_sbm=%.2f, phi=%i, tau=%i", letters[m], this@Rho_a, this@Beta_sbm, this@Phis[m], this@Tau)
				title(main = tmpstring, font.main = 2, cex.main = 1.45)

				for(n in seq_len(this@modules[[m]]@node_num)) {
					x1 <- this@modules[[m]]@nodes[[n]]@weights[[3]]
					y1 <- this@modules[[m]]@nodes[[n]]@weights[[4]]
					x2 <- 1 - this@modules[[m]]@nodes[[n]]@weights[[7]]
					y2 <- 1 - this@modules[[m]]@nodes[[n]]@weights[[8]]
					if(this@modules[[m]]@nodes[[n]]@cluster_ID > 0) {
						rect(x1, y1, x2, y2,col = (this@modules[[m]]@nodes[[n]]@cluster_ID - 1 + ColorShift) %% MaxClusterNum + 1)	
					}
				}
				
				dev.off()
			}
		}
	})	
}
